@@ -53,8 +53,11 @@ class grid(object):
             self.transformers = []
 
         self.lines = data['lines']
-
-
+        self.loads = data['loads']
+        
+        if 'vscs' in data:
+            self.vscs = data['vscs']
+        else: self.vscs = []
 
         
     def dae2vi(self):
@@ -67,9 +70,11 @@ class grid(object):
 
         '''
         n2a = {'1':'a','2':'b','3':'c','4':'n'}
+        a2n = {'a':1,'b':2,'c':3,'n':4}
+        
         V_node_list = []
-        I_node_list = []
-      
+        I_node_list = [0.0]*len(self.nodes_list)
+        self.I_node_list = I_node_list
         for item in self.nodes_list:
             bus_name,phase_name = item.split('.')
             #i = get_i(self.syst,bus_name,phase_name=n2a[phase_name],i_type='phasor',dq_name='ri')
@@ -77,12 +82,8 @@ class grid(object):
             v = get_v(self.syst,bus_name,phase_name=n2a[phase_name],v_type='phasor',dq_name='ri')
             V_node_list += [v]
 
-        for item in self.nodes_list[self.N_v:]:
-            bus_name,phase_name = item.split('.')
-            i = get_i(self.syst,bus_name,phase_name=n2a[phase_name],i_type='phasor',dq_name='ri')
-            I_node_list += [i]
- 
 
+ 
         V_node = np.array(V_node_list).reshape(len(V_node_list),1)
         
         V_known = np.copy(V_node[:self.N_v])
@@ -94,7 +95,44 @@ class grid(object):
         self.I_known = np.array(I_node_list).reshape(len(I_node_list),1)
 
         self.I_node = np.vstack((self.I_unknown,self.I_known))
-        
+ 
+        for load in self.loads:
+            bus_name = load['bus']
+            if load['type'] == '3P+N':
+                for ph in ['a','b','c','n']:
+                    idx = list(self.nodes_list).index(f"{load['bus']}.{a2n[ph]}") 
+                    i_ = get_i(self.syst,'load_' + bus_name,phase_name=ph,i_type='phasor',dq_name='ri')
+                    self.I_node[idx] += i_ 
+
+            if load['type'] == '1P+N':
+                ph = load['bus_nodes'][0]
+                idx = list(self.nodes_list).index(f"{load['bus']}.{ph}") 
+                i_ = get_i(self.syst,'load_' + bus_name,phase_name=n2a[str(ph)],i_type='phasor',dq_name='ri')
+                self.I_node[idx] += i_ 
+                ph = load['bus_nodes'][1]
+                idx = list(self.nodes_list).index(f"{load['bus']}.{ph}") 
+                i_ = get_i(self.syst,'load_' + bus_name,phase_name=n2a[str(ph)],i_type='phasor',dq_name='ri')
+                self.I_node[idx] += i_ 
+
+        for vsc in self.vscs:
+            bus_name = vsc['bus_ac']
+            phases = ['a','b','c','n']
+            if vsc['type'] == 'ac3ph3wvdcq' or vsc['type'] == 'ac3ph3wpq':
+                phases = ['a','b','c']
+
+            for ph in phases:
+                idx = list(self.nodes_list).index(f"{vsc['bus_ac']}.{a2n[ph]}") 
+                i_ = get_i(self.syst,'vsc_' + bus_name,phase_name=ph,i_type='phasor',dq_name='ri')
+                self.I_node[idx] += i_ 
+                
+            if not vsc['type'] == 'ac3ph3wvdcq'  or vsc['type'] == 'ac3ph3wpq':
+                bus_name = vsc['bus_dc']
+                for ph in ['a','n']:
+                    idx = list(self.nodes_list).index(f"{vsc['bus_dc']}.{a2n[ph]}") 
+                    i_ = get_i(self.syst,'vsc_' + bus_name,phase_name=ph,i_type='phasor',dq_name='r')
+                    self.I_node[idx] += i_ 
+                    
+                
         I_lines = self.Y_primitive @ self.A_conect.T @ self.V_node
         self.I_lines = I_lines
         
@@ -821,7 +859,7 @@ def phasor2inst(grid_obj,bus_name,magnitude='v',to_bus='',phases=['a','b','c'],F
     return Times,out
 
 
-def get_voltage(grid_obj,bus_name):
+def get_voltage(grid_obj,bus_name,output='v_an_m'):
     '''
     Get voltage module of a bus.
 
@@ -830,16 +868,35 @@ def get_voltage(grid_obj,bus_name):
     grid_obj : object of pydae class
     bus_name : string
         name of the bus.
-
+    output : string
+         v_an: a phase to neutral voltage phasor (V).
+         v_an_m: a phase to neutral RMS voltage (V).
+         v_abcn_m: a,b and c phases to neutral voltage phasors (V).
+        
     Returns
     -------
     phase-ground voltage module (V).
 
     '''
-    v_a = grid_obj.get_value(f'v_{bus_name}_a_r') + 1j* grid_obj.get_value(f'v_{bus_name}_a_i')
-    U_meas = np.abs(v_a) 
     
-    return U_meas
+    if output in ['v_an','v_bn','v_cn']:
+        v_sub = f'v_{bus_name}_{output[-2]}'
+        v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
+        return v
+    
+    if output in ['v_an_m','v_bn_m','v_cn_m']:
+        v_sub = f'v_{bus_name}_{output[-2]}'[:-2]
+        v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
+        return np.abs(v)
+    
+    if output in ['v_abcn']:
+        v_list = []
+        for ph in ['a','b','c']:
+            v_sub = f'v_{bus_name}_{ph}'
+            v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
+            v_list += [v]
+        return np.array(v_list).reshape((3,1))
+    
 
 
 @numba.njit(cache=True)
@@ -938,7 +995,7 @@ def change_line(system,bus_j,bus_k, *args,**kwagrs):
         R = line['R_pu']*S_base/S_line  # in pu of the system base
         X = line['X_pu']*S_base/S_line  # in pu of the system base
     if 'X' in line:
-        U_base = system.get_value(f'U_{bus_j}_n')
+        U_base = system.get_value(f'U_{bus_j}_n') 
         Z_base = U_base**2/S_base
         R = line['R']/Z_base  # in pu of the system base
         X = line['X']/Z_base  # in pu of the system base
@@ -947,6 +1004,15 @@ def change_line(system,bus_j,bus_k, *args,**kwagrs):
         Z_base = U_base**2/S_base
         R = line['R_km']*line['km']/Z_base  # in pu of the system base
         X = line['X_km']*line['km']/Z_base  # in pu of the system base
+    if 'Bs_km' in line:
+        U_base = system.get_value(f'U_{bus_j}_n')
+        Z_base = U_base**2/S_base
+        print('U_base',U_base,'Z_base',Z_base)
+        Y_base = 1.0/Z_base
+        Bs = line['Bs_km']*line['km']/Y_base  # in pu of the system base
+        bs = Bs
+        system.set_value(f'bs_{line_name}',bs)
+        print(bs)
     G =  R/(R**2+X**2)
     B = -X/(R**2+X**2)
     system.set_value(f"g_{line_name}",G)
@@ -986,3 +1052,71 @@ def get_line_i(system,bus_from,bus_to,U_kV=66e3):
     
     return I_jk
 
+def get_line_s(system,bus_from,bus_to,U_kV=66e3):
+    
+    
+    if f"b_{bus_from}_{bus_to}" in system.params_list: 
+        bus_j = bus_from
+        bus_k = bus_to
+        current_direction = 1.0
+    elif f"b_{bus_to}_{bus_from}" in system.params_list: 
+        bus_j = bus_to
+        bus_k = bus_from
+        current_direction = -1.0
+    else: 
+        print(f'No line from {bus_from} to {bus_to}')
+        return
+    
+    line_name = f"{bus_j}_{bus_k}"
+    V_j_m = system.get_value(f"V_{bus_j}")
+    theta_j = system.get_value(f"theta_{bus_j}")
+    V_k_m = system.get_value(f"V_{bus_k}") 
+    theta_k = system.get_value(f"theta_{bus_k}")
+    
+    V_j = V_j_m*np.exp(1j*theta_j)
+    V_k = V_k_m*np.exp(1j*theta_k)
+    
+    Y_jk = system.get_value(f"g_{line_name}") + 1j*system.get_value(f"b_{line_name}")
+    S_base = system.get_value('S_base')
+    U_base = system.get_value(f"U_{bus_j}_n")
+    I_jk_pu = current_direction*Y_jk*(V_j - V_k)
+    I_base = S_base/(np.sqrt(3)*U_base)
+    I_jk = I_jk_pu*I_base
+    S_jk_pu = V_j*np.conj(I_jk_pu)
+    S_jk = S_base*S_jk_pu
+    
+    return S_jk
+
+
+def set_powers(grid_obj,bus_name,s_cplx,mode='urisi_3ph'):
+    '''
+    Function for simplifying the power setting.
+
+    Parameters
+    ----------
+    grid_obj : pydae object
+        pydae object with grid of type pydgrid.
+    bus_name : string
+        name of the bus to xhange the power.
+    s_cplx : TYPE
+        complex power (negative for load, positive for generated).
+
+    Returns
+    -------
+    None.
+
+    '''
+    
+    if mode == 'urisi_3ph':
+        p = s_cplx.real
+        q = s_cplx.imag
+        for ph in ['a','b','c']:
+            grid_obj.set_value(f'p_{bus_name}_{ph}',p/3)
+            grid_obj.set_value(f'q_{bus_name}_{ph}',q/3)
+        
+    if mode == 'urisi_abc':
+        p = s_cplx.real
+        q = s_cplx.imag
+        for ph in ['a','b','c']:
+            grid_obj.set_value(f'p_{bus_name}_{ph}',p/3)
+            grid_obj.set_value(f'q_{bus_name}_{ph}',q/3)        
\ No newline at end of file
